//! CLI flags module
use super::*;
use std::{
    borrow::Cow,
    any::Any,
    error,
    fmt,
};
use super::error::{
    Error,
    ContextualError,
};
use iterators::*;

type ShortFlag= (char,);

#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Kind<'a>
{
    /// Any number of these long-form arguments starting with `--`, and flag for case-sensitivity.
    /// Can be empty to match just `--`
    Long(&'a [&'a str], bool),
    /// Any number of these short-form (single character) arguments together or apart (starting with `-`).
    /// Can be empty for a single `-` character, and a flag for case-sensitivity.
    Short(&'a [ShortFlag], bool),
    /// A combination of any long + short flags
    Abrev(&'a [&'a str], &'a [ShortFlag], bool, bool),
    /// A single `--`.
    ///
    /// # Note
    /// Shorthand for `Long(&[], _)`
    Terminator,
    /// A single `-`.
    ///
    /// # Note
    /// Shorthand for `Short(&[], _)`
    TerminatorShort,
    /// An exact match with no prefix and flag for case-sensitivity.
    Exact(&'a str, bool),
    /// Match everything.
    Any,
}

impl<'b> Kind<'b>
{
    /// Check if string `value` matches this `Kind`.
    ///
    /// Returns the matched part.
    /// # Note
    /// In this case of `Short` (or `Abrev` short match), return the rest of the argument string including the matched character
    ///
    /// # Short case-sensitivity
    /// Single character matching case sensitivity can produce unexpected results if the lowercase transform of that character yields multiple characters. They are all checked for match.
    pub fn is_match<'a, 'c>(&'a self, value: &'b str) -> Option<&'a str>
    where 'c: 'a,
    'a: 'b
    {
	match self {
	    Self::Long(&[], _) |
	    Self::Abrev(&[], &[], _, _) |
	    Self::Terminator if value == "--" => Some("--"),
	    
 	    Self::Short(&[], _) |
	    Self::TerminatorShort if value == "-" => Some("-"),
	    
	    Self::Abrev(any, _, true, _) |
	    Self::Long(any, true) if value.starts_with("--") => {
		let value = &value[2..];
		
		for &which in any.iter() {
		    if which == value {
			return Some(which);
		    }
		}
		None
	    },
	    Self::Abrev(any, _, false, _) |
	    Self::Long(any, false) if value.starts_with("--") => {
		let value = value[2..].to_lowercase();
		
		for (orig, which) in any.iter().map(|x| (x, x.to_lowercase())) {
		    if which == value {
			return Some(orig);
		    }
		}
		None
	    },

	    Self::Abrev(_, any, _, true) |
	    Self::Short(any, true) if value.starts_with("-") => {
		let mut value = value[1..].chars();
		let iter = &mut value;
		for can_be in iter {
		    if any.contains(&(can_be,)) {
			return Some(value.as_str());
		    }
		}
		None
	    },
	    
	    //Self::Abrev(_, any, _, false) |
	    Self::Short(any, false) if value.starts_with("-") => {
		let mut value = value[1..].chars();
		let iter = &mut value;
		let any: Vec<_> = any.iter().flat_map(|(chr,)| chr.to_lowercase()).collect();
		for can_be in iter.flat_map(|x| x.to_lowercase()) {
		    if any.contains(&can_be) {
			return Some(value.as_str());
		    }
		}
		None
	    },
	    
	    Self::Exact(exact, false) if value.to_lowercase() == exact.to_lowercase() => Some(value),
	    Self::Exact(exact, true) if &value == exact => Some(value),
	    
	    Self::Any => Some(value),
	    _ => None,
	}
    }
    /// Create a reference to this `Kind` with a lower lifetime.
    pub fn as_ref<'a>(&'a self) -> Kind<'a>
    {
	match &self {
	    Self::Abrev(strs, chrs, sl, cl) => Kind::Abrev(strs,chrs, *sl, *cl),
	    Self::Long(strs, bl) => Kind::Long(strs, *bl),
	    Self::Short(chrs, bl) => Kind::Short(chrs, *bl),
	    Self::Terminator => Kind::Terminator,
	    Self::TerminatorShort => Kind::TerminatorShort,
	    Self::Exact(stri, bl) => Kind::Exact(&stri[..], *bl),
	    Self::Any => Kind::Any,
	}
    }

    /// Split `Kind::Abrev` into short and long
    pub fn split_abrev(&self) -> Result<(Kind<'_>, Kind<'_>), &Self>
    {
	match self {
	    Self::Abrev(long, short, lcase, scase) => {
		Ok((Kind::Long(long,*lcase), Kind::Short(short, *scase)))
	    },
	    e => Err(e)
	}
    }
}

/// Helper for `Cow` variants
macro_rules! cow
{
    (& $ex:expr) => (Cow::Borrowed($ex.into()));
    (~ $ex:expr) => (Cow::Owned($ex.into()));
}

impl<'a> fmt::Display for Kind<'a>
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	//TODO: Change for this to write directly to `f` instead of allocating temporary strings and `Cow`ing to it.
	write!(f, "{}", match self {
	    Self::Terminator |
	    Self::Long(&[], _) |
	    Self::Abrev(&[], &[], _, _) => cow!(& "--"),
	    
	    Self::TerminatorShort |
	    Self::Short(&[], _) => cow!(&  "-"),
	    
	    Self::Exact(ex, case) => cow!(~ format!("`{}`{}", ex, if *case {""} else {"i"})),
	    
	    Self::Any => cow!(& "..."),
	    
	    Self::Long(slice, false) |
	    Self::Abrev(slice, &[], false, _)
		=> cow!(~ slice.iter()
			.map(|string| format!("--{}", string))
			.join(", ")),
	    
	    Self::Long(slice, true) |
	    Self::Abrev(slice, &[], true, _)
		=> cow!(~ {
		    let mut string = slice.iter()
			.map(|string| format!("--{}", string))
			.join(", ");
		    string.push_str(" (case sensitive)");
		    string
		}),
	    
	    Self::Short(slice, true) |
	    Self::Abrev(&[], slice, _, true)
		=> cow!(~ std::iter::once('-')
			.chain(slice.iter().map(|x| x.0))
			.collect::<String>()),
	    
	    Self::Short(slice, false) |
	    Self::Abrev(&[], slice, _, false)
		=> cow!(~ {
		    let mut string = std::iter::once('-')
			.chain(slice.iter().map(|x| x.0))
			.collect::<String>();
		    string.push_str(" (case insensitive)");
		    string
		}),
	    
	    Self::Abrev(long, short, lcase, scase) => return write!(f, "{}; {}", Self::Long(long, *lcase), Self::Short(short, *scase)),
	})
    }
}


/// What value(s) (if any) does this argument need
#[derive(Debug, Clone, PartialEq)]
pub enum ValueKind
{
    /// No values, this is a boolean flag
    None,
    /// A value in the form of `--option=value`
    Equals,
    /// The next `n` arguments are the values
    Next(usize),
    /// The whole rest of the iterator is the values
    Rest,
}

impl Default for ValueKind
{
    #[inline]
    fn default() -> Self
    {
	Self::None
    }
}


pub struct ArgState<T=()>
{
    user: T,

    flags: Vec<(Kind<'static>, Cow<'static, str>, ValueKind, Box<(dyn FlagCallback<T, Box<dyn Any + 'static>, eyre::Report> + 'static)>, bool, bool)>,
}

pub trait FlagCallback<U, T,E>: FnMut(&mut StateHandle<'_, U>, usize, &str) -> Result<T, E>{}
impl<U, F, T, E> FlagCallback<U, T,E> for F
where F: FnMut(&mut StateHandle<'_, U>, usize, &str) -> Result<T, E>,
      T: Any + 'static,
      E: Into<eyre::Report> {}

impl<U> ArgState<U>
{
    /// Push a handle to the state
    pub fn push<T: Any + 'static,E: Into<eyre::Report>, F>(&mut self, kind: Kind<'static>, desc: impl Into<Cow<'static, str>>, value: ValueKind, mut handle: F, single: bool) -> usize
    where F: FlagCallback<U, T,E> + 'static
    {
	let handle: Box<dyn FlagCallback<U, Box<(dyn Any + 'static)>, eyre::Report>> =
	    Box::new(move |state, index, value| handle(state, index, value)
		     .map(|x| -> Box<(dyn Any + 'static)> {Box::new(x)})
		     .map_err(Into::into));
	self.flags.push((kind, desc.into(), value, handle, single, true));
	self.flags.len()-1
    }
}



/// A handle to mutate the arg state.
pub struct StateHandle<'a, T>
{
    held: Either<String, Vec<String>>,
    
    state: &'a mut T,
    args: &'a mut (dyn Iterator<Item=String> + 'a),

    chk: &'a mut bool,

    idx: usize,
}

impl<'a, T> StateHandle<'a,T>
{
    /// The user defined state
    pub fn state(&self) -> &T
    {
	self.state
    }
    /// The mutable user defined state 
    pub fn state_mut(&mut self) -> &mut T
    {
	self.state
    }

    /// The current args iterator
    pub fn args(&self) -> &(dyn Iterator<Item=String> + 'a)
    {
	self.args
    }
    /// The mutable current args iterator
    pub fn args_mut(&mut self) -> &mut (dyn Iterator<Item = String> + 'a)
    {
	self.args
    }

    /// Is this callback enabled?
    ///
    /// It will always start as `true`. This can be mutated to `false` to disable further checks for this argument.
    pub fn enabled_mut(&mut self) -> &mut bool
    {
	self.chk
    }
    /// Is this callback enabled?
    ///
    /// This should always be `true`.
    pub fn enabled(&mut self) -> bool
    {
	*self.chk
    }
    /// The index of the current arg
    pub fn index(&self) -> usize
    {
	self.idx
    }

    /// The held values taken from the argument iterator as specified by the argument's definition.
    ///
    /// # Note
    /// This moves the values, if you want to call this more than once, use `clone_held`.
    /// If called more than once will yield no values
    pub fn extract_held(&mut self) -> EitherIter<String, std::vec::IntoIter<String>>
    {
	self.held.take().into_iter()
    }

    /// Clone the held values taken from the argument iterator for this invokation.
    ///
    /// # Note
    /// This clones all the values, instead of moving like `extract_held`, so multiple calls to this will yield the same clones.
    /// However, if `extract_held` has been called, this will yield no values.
    pub fn clone_held(&self) -> EitherIter<String, std::vec::IntoIter<String>>
    {
	self.held.clone().into_iter()
    }
}


impl<U> ArgState<U>
{
    /// Parse this argument iterator.
    ///
    /// # Note
    /// This mutates the argument state container indefinately, and multiple calls to it will keep the mutated state.
    /// So this might not behave as expected (e.g. callbacks marked `single` that were fired in the first call will not be fired in the second, etc.)
    pub fn parse<I,T: Into<String>>(&mut self, input: I) -> Result<Vec<Box<dyn Any+'static>>, ContextualError>
    where I: Iterator<Item=T>
    {
	let mut output = Vec::with_capacity(self.flags.len());
	let mut input = input.map(Into::into).fuse();

	let mut i=0;
	while let Some(arg) = input.next()
	{
	    for (kind, desc, value, callback, single, chk) in self.flags.iter_mut().filter(|(_,_,_,_,_,chk)| *chk)
	    {
		if *single {
		    *chk = false;
		}
		//TODO: When value is `Equals`, split at `=` and pass only the first part to `is_match`.
		if let Some(kind_match) = kind.is_match(&arg[..]) {
		    //TODO: Check `value` matches `input` or `kind_match`
		    output.push(callback(&mut StateHandle{
			state: &mut self.user,
			args: &mut input,
			chk,
			idx: i,
			held: Either::None, //TODO: This will be for either one `=<value>` or many `<values...>` from args as according to `value`.
		    }, i, kind_match)
				.map_err(|err| Error{flag: kind.clone(), from: err, desc: Some(desc.clone().into_owned()), arg: (i, arg.clone())})?);
		}
	    }
	    i+=1;
	}

	Ok(output)
    }

    /// Consume into the user-provided state value
    #[inline] pub fn into_inner(self) -> U
    {
	self.user
    }
}


#[cfg(test)]
mod tests
{
    use super::*;
    #[test]
    fn errors() -> Result<(), eyre::Report>
    {
	color_eyre::install()?;
	
	let err= Error {
	    flag: Kind::Abrev(&["test", "real-test", "more-tests"], &[('t',), ('R',)], false, true),
	    arg: (0, "owo".to_owned()),
	    desc: Some("expected between 1-10".to_string()),
	    from: eyre!("Invalid number").wrap_err("Parsing failure"),
	};
	println!("No ctx: `{} {:?}`\n", err, err);
	let err = err.with_context();
	println!("With ctx: `{} {:?}`\n", err, err);

	//return Err(err)?;
	Ok(())
    }
}
